@@ -505,6 +505,7 @@ let test_Z() =
   Printf.printf "sign -1\n = %i\n" (I.sign I.minus_one);
   Printf.printf "sign 2^300\n = %i\n" (I.sign p300);
   Printf.printf "sign -2^300\n = %i\n" (I.sign (I.neg p300));
+  Printf.printf "gcd 0 0\n = %a\n" pr (I.gcd I.zero I.zero);
   Printf.printf "gcd 0 -137\n = %a\n" pr (I.gcd (I.of_int 0) (I.of_int (-137)));
   Printf.printf "gcd 12 27\n = %a\n" pr (I.gcd (I.of_int 12) (I.of_int 27));
   Printf.printf "gcd 27 12\n = %a\n" pr (I.gcd (I.of_int 27) (I.of_int 12));
@@ -515,6 +516,13 @@ let test_Z() =
   Printf.printf "gcd 0 2^300\n = %a\n" pr (I.gcd (I.of_int 0) p300);
   Printf.printf "gcd 2^120 2^300\n = %a\n" pr (I.gcd p120 p300);
   Printf.printf "gcd 2^300 2^120\n = %a\n" pr (I.gcd p300 p120);
+  Printf.printf "gcd 0 -2^300\n = %a\n" pr (I.gcd (I.of_int 0) (I.neg p300));
+  Printf.printf "gcd 2^120 -2^300\n = %a\n" pr (I.gcd p120 (I.neg p300));
+  Printf.printf "gcd 2^300 -2^120\n = %a\n" pr (I.gcd p300 (I.neg p120));
+  Printf.printf "gcd -2^120 2^300\n = %a\n" pr (I.gcd (I.neg p120) p300);
+  Printf.printf "gcd -2^300 2^120\n = %a\n" pr (I.gcd (I.neg p300) p120);
+  Printf.printf "gcd -2^120 -2^300\n = %a\n" pr (I.gcd (I.neg p120) (I.neg p300));
+  Printf.printf "gcd -2^300 -2^120\n = %a\n" pr (I.gcd (I.neg p300) (I.neg p120));
   Printf.printf "gcdext 12 27\n = %a\n" pr3 (I.gcdext (I.of_int 12) (I.of_int 27));
   Printf.printf "gcdext 27 12\n = %a\n" pr3 (I.gcdext (I.of_int 27) (I.of_int 12));
   Printf.printf "gcdext 27 27\n = %a\n" pr3 (I.gcdext (I.of_int 27) (I.of_int 27));
@@ -523,6 +531,30 @@ let test_Z() =
   Printf.printf "gcdext -12 -27\n = %a\n" pr3 (I.gcdext (I.of_int (-12)) (I.of_int (-27)));
   Printf.printf "gcdext 2^120 2^300\n = %a\n" pr3 (I.gcdext p120 p300);
   Printf.printf "gcdext 2^300 2^120\n = %a\n" pr3 (I.gcdext p300 p120);
+  Printf.printf "gcdext 12 0\n = %a\n" pr3 (I.gcdext (I.of_int 12) I.zero);
+  Printf.printf "gcdext 0 27\n = %a\n" pr3 (I.gcdext I.zero (I.of_int 27));
+  Printf.printf "gcdext -12 0\n = %a\n" pr3 (I.gcdext (I.of_int (-12)) I.zero);
+  Printf.printf "gcdext 0 -27\n = %a\n" pr3 (I.gcdext I.zero (I.of_int (-27)));
+  Printf.printf "gcdext 2^120 0\n = %a\n" pr3 (I.gcdext p120 I.zero);
+  Printf.printf "gcdext 0 2^300\n = %a\n" pr3 (I.gcdext I.zero p300);
+  Printf.printf "gcdext -2^120 0\n = %a\n" pr3 (I.gcdext (I.neg p120) I.zero);
+  Printf.printf "gcdext 0 -2^300\n = %a\n" pr3 (I.gcdext I.zero (I.neg p300));
+  Printf.printf "gcdext 0 0\n = %a\n" pr3 (I.gcdext I.zero I.zero);
+  Printf.printf "lcm 0 0 = %a\n" pr (I.lcm I.zero I.zero);
+  Printf.printf "lcm 10 12 = %a\n" pr (I.lcm (I.of_int 10) (I.of_int 12));
+  Printf.printf "lcm -10 12 = %a\n" pr (I.lcm (I.of_int (-10)) (I.of_int 12));
+  Printf.printf "lcm 10 -12 = %a\n" pr (I.lcm (I.of_int 10) (I.of_int (-12)));
+  Printf.printf "lcm -10 -12 = %a\n" pr (I.lcm (I.of_int (-10)) (I.of_int (-12)));
+  Printf.printf "lcm 0 12 = %a\n" pr (I.lcm I.zero (I.of_int 12));
+  Printf.printf "lcm 0 -12 = %a\n" pr (I.lcm I.zero (I.of_int (-12)));
+  Printf.printf "lcm 10 0 = %a\n" pr (I.lcm (I.of_int 10) I.zero);
+  Printf.printf "lcm -10 0 = %a\n" pr (I.lcm (I.of_int (-10)) I.zero);
+  Printf.printf "lcm 2^120 2^300 = %a\n" pr (I.lcm p120 p300);
+  Printf.printf "lcm 2^120 -2^300 = %a\n" pr (I.lcm p120 (I.neg p300));
+  Printf.printf "lcm -2^120 2^300 = %a\n" pr (I.lcm (I.neg p120) p300);
+  Printf.printf "lcm -2^120 -2^300 = %a\n" pr (I.lcm (I.neg p120) (I.neg p300));
+  Printf.printf "lcm 2^120 0 = %a\n" pr (I.lcm p120 I.zero);
+  Printf.printf "lcm -2^120 0 = %a\n" pr (I.lcm (I.neg p120) I.zero);
   Printf.printf "is_odd 0\n = %b\n" (I.is_odd (Z.of_int 0));
   Printf.printf "is_odd 1\n = %b\n" (I.is_odd (Z.of_int 1));
   Printf.printf "is_odd 2\n = %b\n" (I.is_odd (Z.of_int 2));
